// NOTE: The content of the following header applies to this file as well.

header {
/*
 * OpenFDL Compiler
 * Copyright (C) 2005 Matt Acosta
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */
}

//////////////////////////////////////////////////
// TREEPARSER
//////////////////////////////////////////////////

{
import org.openfdl.core.FrameAttrNode;

/**
 * FDLTreeWalker.java
 * 
 * @author Matt Acosta
 * @version 1.00
 */
}

class FDLTreeParser extends TreeParser;

{
	// Keep track of where we started
	private FrameAttrNode rootNode = null;
	
	// Used to create/add nodes
	private FrameAttrNode posNode;
	private FrameAttrNode tempNode = new FrameAttrNode();
	
	// Slot numbers, tricky (default=0x10000)
	private int tempNum, lastNum = 65535;
	
	private String tempStr;
	
	private boolean errorFound = false;
	
	public FrameAttrNode getRootNode()
	{
		return rootNode;
	}
}

// ENTER TREEPARSER HERE
parse
  : (fdlLine {
  		if (!errorFound)
  		{
	  		// Set slot numbers (we increment to default if missing)
	  		if (tempNum == 0)
	 	 		lastNum++;
	 	 	else
	 	 		lastNum = tempNum;
	 	 	
	 	 	tempNode.setSlotNum(lastNum);
	  		
	  		// Reset fields
	  		tempNum = 0;
	  		tempStr = null;
	  		
	  		// Add to the nodes..
	  		if (rootNode == null)
	  			posNode = rootNode = tempNode;
	  		else
	  			posNode = posNode.appendAttrNode(tempNode);
  		}
  		
  		tempNode = new FrameAttrNode();
  		
  		errorFound = false;
    })*
  ;

// Declarations
fdlLine
  : varDecl
  | funcDecl
  | appFuncDecl
  ;

varDecl
  : attrType:varType {
  		tempNode.setAttrType(attrType.getType());
  	}
    (slotNum:INTLIT {
    	tempNum = Integer.parseInt(slotNum.getText());
    })?
    attrName:IDENT {
    	tempNode.setAttrName(attrName.getText());
    }
    COLON {tempStr = "";}
    typeSpec {
    	tempNode.setVarType(tempStr.trim());
    }
    SEMI
  ;
  exception
  catch [RecognitionException e] {
  	errorFound = true;
  	reportError(e.getMessage());
  }

funcDecl
  : attrType:funcType {
  		tempNode.setAttrType(attrType.getType());
    }
    (slotNum:INTLIT {
    	tempNum = Integer.parseInt(slotNum.getText());
    })?
    attrName:IDENT {
    	tempNode.setAttrName(attrName.getText());
    }
    funcParameters
    COLON {tempStr = "";}
    typeSpec {
    	tempNode.setVarType(tempStr.trim());
    }
    SEMI
  ;
  exception
  catch [RecognitionException e] {
  	errorFound = true;
  	reportError(e.getMessage());
  }

appFuncDecl
  : attrType:appFuncType {
  		tempNode.setAttrType(attrType.getType());
    }
    (slotNum:INTLIT {
    	tempNum = Integer.parseInt(slotNum.getText());
    })?
    attrName:IDENT {
    	tempNode.setAttrName(attrName.getText());
    }
    appFuncParameters
    COLON {tempStr = "";}
    typeSpec {
    	tempNode.setVarType(tempStr.trim());
    }
    SEMI
  ;
  exception
  catch [RecognitionException e] {
  	errorFound = true;
  	reportError(e.getMessage());
  }

// Frame types
varType
  : "var"
  | "appvar"
  ;

funcType
  : "func"
  ;

appFuncType
  : "appfunc"
  ;

// Function parameter list
funcParameters
  : LPAREN!
    pFrameStr:funcReqParam {
    	tempStr = pFrameStr.getText();
    }
    (comma:COMMA {
    	tempStr += comma.getText();
    }
    typeSpec)*
    RPAREN!  { tempNode.setParams(tempStr); }
  ;
  exception
  catch [RecognitionException e] {
  	errorFound = true;
  	reportError(e.getMessage());
  }

appFuncParameters
  : LPAREN!
    appid:appFuncReqParam {
    	tempStr = appid.getText();
    }
    (comma:COMMA {
    	tempStr += comma.getText();
    }
    typeSpec)*
    RPAREN! { tempNode.setParams(tempStr); }
  ;
  exception
  catch [RecognitionException e] {
  	errorFound = true;
  	reportError(e.getMessage());
  }

// Required parameters
funcReqParam
  : "pFrame"
  ;
  exception		// Missing parameter
  catch [RecognitionException e] {
  	errorFound = true;
  	reportError(e.getMessage());
  }

appFuncReqParam
  : "AppID"
  ;
  exception		// Missing parameter
  catch [RecognitionException e] {
  	errorFound = true;
  	reportError(e.getMessage());
  }

// Types

// Known issues:
// * Modifier "const" must be declarted first

typeSpec
  : (attr:typeAttr {
  		tempStr += " " + attr.getText();
  	})?
    typeName
    (ptr:TIMES {
    	tempStr += " " + ptr.getText();
    })?
  ;
  exception
  catch [RecognitionException e] {
  	errorFound = true;
  	reportError("invalid type");
  }

typeName
  : ((dType:dataType { tempStr += " " + dType.getText(); })?
  	  name:IDENT { tempStr += " " + name.getText(); }
  	)
  | ((mod:typeMod { tempStr += " " + mod.getText(); })?
	  nVal:numericType { tempStr += " " + nVal.getText(); }
	)
  | dVal:decimalType { tempStr += " " + dVal.getText(); }
  ;

numericType
  : "char"
  | "int"
  | "long"
  | "short"
  ;

decimalType
  : "float"
  | "double"
  ;

dataType
  : "struct"
  | "union"
  ;

typeAttr
  : "const"
  ;

typeMod
  : "unsigned"
  | "signed"
  ;

//////////////////////////////////////////////////
// PARSER
//////////////////////////////////////////////////

{
/**
 * FDLParser.java
 * 
 * @author Matt Acosta
 * @version 1.00
 */
}

class FDLParser extends Parser;

options {
  defaultErrorHandler=true;
  buildAST=true;
}

{
	private boolean parseFail = false;
}

// ENTER PARSER HERE

// Known issues:
// * See "Frame types", EOF! will throw an error, but with consequences..

parse
  : (fdlLine)* {
  	if (parseFail)
  		reportWarning("parser failure before EOF");
    }
  ;

// Declarations
fdlLine
  : varDecl
  | funcDecl
  | appFuncDecl
  ;

varDecl
  : varType (INTLIT)? IDENT COLON typeSpec SEMI
  ;
  exception		// missing colons, semicolons, or bad identifiers
  catch [RecognitionException e] {
  	parseFail = true;
  	reportError("line " + e.getLine() + ":" + e.getColumn() +
  		", " + e.getMessage());
  }

funcDecl
  : funcType (INTLIT)? IDENT funcParameters COLON typeSpec SEMI
  ;
  exception		// missing colons, semicolons, or bad identifiers
  catch [RecognitionException e] {
  	parseFail = true;
  	reportError("line " + e.getLine() + ":" + e.getColumn() +
  		", " + e.getMessage());
  }

appFuncDecl
  : appFuncType (INTLIT)? IDENT appFuncParameters COLON typeSpec SEMI
  ;
  exception		//	missing colons, semicolons, or bad identifiers
  catch [RecognitionException e] {
  	parseFail = true;
  	reportError("line " + e.getLine() + ":" + e.getColumn() +
  		", " + e.getMessage());
  }

// Function parameter list
funcParameters
  : LPAREN funcReqParam (COMMA typeSpec)* RPAREN
  ;
  exception		// missing parenthesis or commas or invalid parameters
  catch [RecognitionException e] {
  	reportError("line " + e.getLine() + ":" + e.getColumn() +
  		", " + e.getMessage());
  }

appFuncParameters
  : LPAREN appFuncReqParam (COMMA typeSpec)* RPAREN
  ;
  exception		// missing parenthesis or commas or invalid parameters
  catch [RecognitionException e] {
  	reportError("line " + e.getLine() + ":" + e.getColumn() +
  		", " + e.getMessage());
  }

// Frame types

// Known issues:
// * Misspelled type does not throw error, exits parser

varType
  : "var"
  | "appvar"
  ;
  exception
  catch [RecognitionException e] {
  	reportError("missing attribute type, line " + e.getLine() + ":" +
  		e.getColumn() + ", " + e.getMessage());
  }

funcType
  : "func"
  ;
  exception
  catch [RecognitionException e] {
  	reportError("missing attribute type, line " + e.getLine() + ":" +
  		e.getColumn() + ", " + e.getMessage());
  }

appFuncType
  : "appfunc"
  ;
  exception
  catch [RecognitionException e] {
  	reportError("missing attribute type, line " + e.getLine() + ":" +
  		e.getColumn() + ", " + e.getMessage());
  }

// Required parameters
funcReqParam
  : "pFrame"
  ;
  exception
  catch [RecognitionException e] {
  	reportError("missing required parameter, line " + e.getLine() + ":" +
  		e.getColumn() + ", " + e.getMessage());
  }

appFuncReqParam
  : "AppID"
  ;
  exception
  catch [RecognitionException e] {
  	reportError("missing required parameter, line " + e.getLine() + ":" +
  		e.getColumn() + ", " + e.getMessage());
  }

// Types

// Known issues:
// * Modifier "const" must be declarted first

typeSpec
  : (typeAttr)? typeName (TIMES)?
  ;
  exception
  catch [RecognitionException e] {
  	reportError("invalid type, line " + e.getLine() + ":" +
  		e.getColumn() + ", " + e.getMessage());
  }

typeName
  : ((dataType)? IDENT)
  | ((typeMod)? numericType)
  | decimalType
  ;

numericType
  : "char"
  | "int"
  | "long"
  | "short"
  ;

decimalType
  : "float"
  | "double"
  ;

dataType
  : "struct"
  | "union"
  ;

typeAttr
  : "const"
  ;

typeMod
  : "unsigned"
  | "signed"
  ;

//////////////////////////////////////////////////
// SCANNER
//////////////////////////////////////////////////

{
/**
 * FDLScanner.java
 * 
 * @author Matt Acosta
 * @version 1.00
 */
}

class FDLScanner extends Lexer;

options {
  charVocabulary = '\0'..'\377';
  testLiterals=false;       // Don't auto test literals
  k=2;                      // two character lookahead
}

// Whitespace (ignored)
WS
  :
    ( ' '
   	| '\t'
   	| '\f'
   	
   	// Handle newlines
   	| ( "\r\n"	// Windows
   	  | '\r'	// Macs
   	  | '\n'	// UNIX
   	  )
   	  
   	  // Increment scanner line count
   	  { newline(); }
   	)
   	{$setType(Token.SKIP);}
  ;

// Single line comments
COMMENT
  : "#" (~('\n'|'\r'))*
    {$setType(Token.SKIP);}
  ;

// Literals
INTLIT
  : ('0'..'9')+
  ;

// Identifiers
IDENT
  options {testLiterals=true;}
  : ('a'..'z'|'A'..'Z'|'_') ('a'..'z'|'A'..'Z'|'0'..'9'|'_')*
  ;

// Opertators
COLON
  options {paraphrase=":";}
  :':' ;
SEMI
  options {paraphrase=";";}
  : ';' ;
COMMA
  options {paraphrase=",";}
  : ',' ;
LPAREN 
  options {paraphrase="(";}
  : '(' ;
RPAREN
  options {paraphrase=")";}
  : ')' ;
TIMES
  options {paraphrase="*";}
  : '*' ;
LBRAKET : '[' ;
RBRAKET : ']' ;
